/*
 * @(#)dog.java    Jan 1, 1998
 * 
 * Copyright (c) 1998, Particle
 * 
 * Sale of this "Software" is prohibited!  Only FREE
 * distribution is alowed.
 * 
 *
 * -- Particle --
 */

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.lang.Double;

import pQueue;
import pPoint2D;
import pLine2D;
import pPlane2D;
import pBSPTree;

/**
 * A <code>dog</code> class simply uses the <code>pBSPTree</code>
 * class and renders a bunch of lines in a way, giving the appearance
 * of a 3D room.  (when in reality it's just 2D)
 * <p>
 * This class must accept a <code>pQueue</code> object containing
 * lines, which it later inserts into the BSP, which sorts them, 
 * so that this <code>dog</code> object can render them.
 * <p>
 * I call it "Dog 3D"!
 *
 * @version     1.0 1/2/1998
 * @author      Particle
 */
public class dog extends Frame implements WindowListener,KeyListener{
	
	/**
	 * the BSP tree object 
	 */
	public pBSPTree myBSP = new pBSPTree();
	
	/** 
	 * a queue used to insert stuff onto tree, and
	 * retrieve sorted lines 
	 */
	public pQueue myList = new pQueue();

	/**
	 * the initial eye point, (or the position of a "player")
	 */
	public pPoint2D eye = new pPoint2D(220,220);

	/**
	 * the eye point angle... (or the "direction" the "player" is facing)
	 */
	public double pangle = 0;

	/**
	 * the drawable area width 
	 */
	int screen_width = 320;

	/**
	 * the drawable area height
	 */
	int screen_height = 240;

	/**
	 * a buffered image, used for double buffering.
	 */
	BufferedImage double_buffer = null;

	/**
	 * a graphics contect for the double buffer, 
	 * (I think I am getting lazy... ) 
	 * this approach is VERY slow 
	 */
	Graphics double_graphics = null;

	/**
	 * method to setup the double buffer stuff...
	 */
	public void setup()
	{
		double_buffer = new BufferedImage(screen_width,
			screen_height,BufferedImage.TYPE_INT_ARGB);
		double_graphics = double_buffer.createGraphics();
	}

	/**
	 * Constructs and initializes some stuff, accepts a <code>pQueue</code>
	 * of lines, and inserts them onto the BSP Tree... 
	 */
	public dog(pQueue lines)
	{
		super("Dog 3D");
		setup();
		setSize(screen_width+100,screen_height+100);
		myList = lines;
		myBSP.pBuildBSPTree(myList);
		addWindowListener(this);
		addKeyListener(this);
		setResizable(false);	/* can't resize window */ 
		show();
		repaint();
	}
	
	/**
	 * given a line, and a graphics context, 
	 * (and a bunch of class variables), it transforms, 
	 * rotates the line coordinates, and draws a 
	 * 3D type version of the line onto the graphics
	 * context.
	 *
 	 * NOTE: VERY VERY Buggy! (Needs LOTS of improvement)
	 */
	void _mydrawPoly(Graphics g,pLine2D tmp){
      	double x1=tmp.getEnd().getx();
        	double y1=tmp.getEnd().gety(); 
        	double x2=tmp.getStart().getx();
        	double y2=tmp.getStart().gety();
        	double pCos = Math.cos(pangle);
        	double pSin = Math.sin(pangle);

        	int [] x = new int[4];
        	int [] y = new int[4];

        	double pD=-pSin*eye.getx()+pCos*eye.gety();      
        	double pDp=pCos*eye.getx()+pSin*eye.gety();
       
        	double rz1,rz2,rx1,rx2;

        	int Screen_x1=0,Screen_x2=0;
		double Screen_y1,Screen_y2,Screen_y3,Screen_y4;

        	rz1=pCos*x1+pSin*y1-pDp;     //perpendicular line to the players
        	rz2=pCos*x2+pSin*y2-pDp;     //view point

        	if((rz1<1) && (rz2<1)){
         		return;
        	}
		
		rx1=pCos*y1-pSin*x1-pD;
        	rx2=pCos*y2-pSin*x2-pD;   
		
		double pTan = 0;
        	if((x2-x1) == 0){
			pTan = Double.MAX_VALUE;
        	}else{
			pTan = (y2-y1)/(x2-x1);
        	}  

        	pTan = (pTan-Math.tan(pangle))/(1+(pTan*Math.tan(pangle)));

        	if(rz1 < 1){
			rx1+=(1-rz1)*pTan;
			rz1=1;
        	}
        	if(rz2 < 1){
			rx2+=(1-rz2)*pTan;
			rz2=1;
        	}
		
		double z1 = screen_width/2/rz1;
        	double z2 = screen_width/2/rz2;

		//get the pespective X coordinates on the screen...
        	Screen_x1=(int)(screen_width/2-rx1*z1);
        	Screen_x2=(int)(screen_width/2-rx2*z2);

        	//cliping...(can the wall be seen?)
        	if(Screen_x1>screen_width)     return;      //can not be seen...
		if(Screen_x2<0)	return;

        	//set the height...
		int ph=40;
        	int wt=88;
        	int wb=-40;

		Screen_y1=(double)screen_height/2-(double)wt*z1;
		Screen_y4=(double)screen_height/2-(double)wb*z1;
		Screen_y2=(double)screen_height/2-(double)wt*z2;
		Screen_y3=(double)screen_height/2-(double)wb*z2;

		if(Screen_x1<0){
			Screen_y1+=(0-Screen_x1)*(Screen_y2-Screen_y1)/
					(Screen_x2-Screen_x1);
			Screen_y4+=(0-Screen_x1)*(Screen_y3-Screen_y4)/
					(Screen_x2-Screen_x1);
			Screen_x1=0;
		}
		if(Screen_x2>(screen_width)){
			Screen_y2-=(Screen_x2-screen_width)*
					(Screen_y2-Screen_y1)/
					(Screen_x2-Screen_x1);
			Screen_y3-=(Screen_x2-screen_width)*
					(Screen_y3-Screen_y4)/
					(Screen_x2-Screen_x1);
			Screen_x2=screen_width;
		}

        	if((Screen_x2-Screen_x1)==0)  return;

        	x[0] = (int)Screen_x1;
        	y[0] = (int)(Screen_y1);
        	x[1] = (int)Screen_x2;
        	y[1] = (int)(Screen_y2);
        	x[2] = (int)Screen_x2;
        	y[2] = (int)(Screen_y3);
        	x[3] = (int)Screen_x1;
        	y[3] = (int)(Screen_y4);
        	g.fillPolygon(x,y,4);
	}




	/**
	 * the method to get the sorted lines out of a BSP Tree, and 
	 * call a function to render those lines in the order recieved
	 * from the BSP Tree... 
	 */
	void pRender(Graphics g){
		int i = 0;
		pQueue q = myBSP.pGetSortedLines(eye);
		g.setColor(Color.black);
		g.fillRect(0,0,screen_width,screen_height);
		pLine2D tmp;
		while(!q.empty()){
			i++;
			tmp = (pLine2D)q.remove();
				/* back face removal */
			if(tmp.getPlane().eval(eye) == pPlane2D.IN_BACK)
				continue;
				/* just pick a unique color for this "line" */
			g.setColor(new Color(i*1024+0x555555));
			_mydrawPoly(g,tmp);
   		}
	}

	/**
	 * overriding standard "paint" method
	 */
	public void paint(Graphics g){
		pRender(double_graphics);
		g.drawImage(double_buffer,50,50,this);
   	}

	/**
	 * overriding standard "update" method
	 */
	public void update(Graphics g){
		paint(g);
	}

	/**
	 * catch window closing, and dispose of it. 
	 */
	public void windowClosing(WindowEvent e)
	{
		e.getWindow().dispose();
		System.exit(0);
	}
      
   	public void windowOpened(WindowEvent e){}
   	public void windowClosed(WindowEvent e){}
    	public void windowIconified(WindowEvent e){}
    	public void windowDeiconified(WindowEvent e){}
    	public void windowActivated(WindowEvent e){}
    	public void windowDeactivated(WindowEvent e){}

	public void keyTyped(KeyEvent e){}

	/**
	 * get some primitive user input
	 * (the arrow keys) 
	 */
	public void keyPressed(KeyEvent e)
	{	
		if(e.getKeyCode() == KeyEvent.VK_UP){
            	eye.v[0] += Math.cos(pangle)*7;
              	eye.v[1] += Math.sin(pangle)*7;
        	}else if(e.getKeyCode() == KeyEvent.VK_DOWN){
            	eye.v[0] -= Math.cos(pangle)*7;
              	eye.v[1] -= Math.sin(pangle)*7;
        	}else if(e.getKeyCode() == KeyEvent.VK_LEFT){
              	pangle += 2*Math.PI/64;
        	}else if(e.getKeyCode() == KeyEvent.VK_RIGHT){
            	pangle -= 2*Math.PI/64;
        	}
		repaint();
	}

	public void keyReleased(KeyEvent e){} 
}
