/*
 * @(#)pBSPNode.java    Dec 31, 1997
 * 
 * Copyright (c) 1997, Particle
 * 
 * Sale of this "Software" is prohibited!  Only FREE
 * distribution is alowed.
 *
 * -- Particle --
 */

import pPlane2D;
import pQueue;

/**
 * A <code>pBSPNode</code> class is a very basic Node with 2 children.
 * <p>
 * This is basically a plain vanilla node, only it has 2 children, making
 * it ideal for a binary tree.
 * <p> 
 * The data on the node is a <code>pQueue</code> object, (kinda like a list)
 * of stuff... and a partition plane. (a <code>pPlane2D</code> object)
 * 
 * @version     1.0 12/31/1997
 * @author      Particle
 */

public class pBSPNode{

	/**
	 * the partion plane, (hey, it IS a BSP node...)
	 */
	public pPlane2D partition = null;
  	
	/**
	 * a queue of "lines" that are on the partition plane.  This can also
	 * be a "list" (instead of a queue), I just like the queue better, and
	 * besides, that's how I return the "sorted" ones, so, why not use it...
	 * <p>
	 * although, a simple List approach may be a bit faster... 
	 */
	public pQueue lines = null;
  	
	/**
	 * the "next" node pointing to the "front" child
	 */
	public pBSPNode front = null;
 		
	/**
	 * the "next" node pointing to the "back" child
	 */
	public pBSPNode back = null;

	/** 
	 * Constructs a pBSPNode object, with default settings.
	 */
	public pBSPNode(){
    		partition = new pPlane2D();
       	lines = new pQueue();
	}
	
	/**
	 * a method to get the "back" child node... 
	 * @return the back child
	 */
 	public pBSPNode getback(){
       	return back;
	}
	
	/**
	 * a method to get the "front" child node...
	 * @return the front child
	 */
 	public pBSPNode getfront(){
      	return front;
	}

	/**
	 * a method to set the back child node to a new value
	 * @param l the node to set "back" child to... 
	 */
	public void setback(pBSPNode l){
      	back = l;
	}

	/**
	 * a method to set the front child node to a new value
	 * @param r the node to set "front" child to...
	 */
	public void setfront(pBSPNode r){
      	front = r;
	}

	/**
	 * a standard <code>toString</code> method
	 * <p>
	 * Note that this could easily be modified to "return" the whole 
	 * tree structure... 
	 *
	 * @return just the data in this node.
	 */
	public String toString(){
		return new String(""+partition+"/"+lines);
	}
}
