/*
 * @(#)pBSPTree.java    Dec 31, 1997
 * 
 * Copyright (c) 1997, Particle
 * 
 * Sale of this "Software" is prohibited!  Only FREE
 * distribution is alowed.
 *
 * -- Particle --
 */

import pQueue;
import pBSPNode;
import pLine2D;
import pPlane2D;

/**
 * A <code>pBSPTree</code> class is a Binary Space Partition Tree!
 * <p>
 * This method mostly works with Queues, inserting entire structures,
 * and sorting and returning entire structures as well... 
 * 
 * @version     1.0 12/31/1997
 * @author      Particle
 */

public class pBSPTree{

	/**
	 * The root of the TREE. (this node can take you anywhere.)
	 */
	public pBSPNode root;

	/**
	 * the list of sorted lines, it's here, to speed things up,
	 * and eliminate it's passing as a parameter to the sorting 
	 * function
 	 */
	pQueue list;
      
	/**
	 * the current eye position, it's passed to the sorting 
	 * function, which later sets this variable, so that 
	 * the recursive method won't have to pass it as a 
	 * parameter when it's sorting... 
	 */
	pPoint2D eye;


	/** 
	 * Constructs a <code>pBSPTree</code> object, and initialized root to 
	 * <code>null</code>
	 */
 	public pBSPTree()
	{
      	root = null;
    	}

	/**
	 * a recurcive (& private) function to build a bsp 
	 * tree.  Users should be calling a "friendly" function
	 * below... 
	 * 
	 * @param tree the current node of the tree.
	 * @param lines a pQueue of lines to examine.
	 */
     	void pBuildBSPTree(pBSPNode tree, pQueue lines)
   	{
    		pLine2D current_line = (pLine2D)lines.remove();
       	tree.partition = current_line.getPlane();
		tree.lines.insert(current_line);
		pQueue front_list = new pQueue();
		pQueue back_list = new pQueue();
		pLine2D line = null;
		while(!lines.empty()){
			line = (pLine2D)lines.remove();
			int result = tree.partition.eval(line);
			if(result==pPlane2D.IN_FRONT){         /* in front */
				front_list.insert(line);
            	}else if(result==pPlane2D.IN_BACK){   /* in back */
            		back_list.insert(line);
           		}else if(result==pPlane2D.SPANNING){          /* spanning */
              		pLine2D [] split_line = null;
               		split_line = tree.partition.split(line);
                		if(split_line != null){
                     		front_list.insert(split_line[0]);
                     		back_list.insert(split_line[1]);
                     	}else{
                        	/* error here, (TODO: throw some Exception) */
                   	}
            	}else if(result==pPlane2D.COINCIDENT){
                                tree.lines.insert(line);
               	}
      	}

      	if(!front_list.empty()){
            	tree.front = new pBSPNode();
            	pBuildBSPTree(tree.front,front_list);
        	}
       	if(!back_list.empty()){
            	tree.back = new pBSPNode();
               	pBuildBSPTree(tree.back,back_list);
       	}
  	}

   	/**
	 * a 'friendly' function for the user to construct a 
	 * BSP Tree.
	 * <p> 
	 * The pQueue parameter should have ALL the lines
	 * which are to be inserted, if something is on the tree
	 * previous to that, it will be overwritten by this new
	 * insertion... 
	 * 
	 * @param lines a pQueue of lines to insert into the BSP Tree. 
	 * @see pQueue
	 */
  	public void pBuildBSPTree(pQueue lines)
  	{
      	root = new pBSPNode();
        	pBuildBSPTree(root,lines);
  	}

	/**
	 * a function to get the sorted lines in relation 
	 * to the <code>this.eye</code>, the resulting sorted
	 * list is stored in <code>this.list</code>
	 * <p>
	 * The user should call a "user friendly" function below.
	 *
	 * @param tree the currecnt pBSPNode in recursion.
	 */
  	void pGetSortedLines(pBSPNode tree)
  	{
      	pLine2D tmp;
   		if(tree == null)        
			return; /* check for end */
     		int result = tree.partition.eval(eye);
        	if(result == pPlane2D.IN_FRONT){
            	pGetSortedLines(tree.back);
             	for(int i = 0;i<tree.lines.number();i++)
				list.insert(tree.lines.peek(i));
			pGetSortedLines(tree.front);
         	}else if(result == pPlane2D.IN_BACK){
            	pGetSortedLines(tree.front);
               	for(int i = 0;i<tree.lines.number();i++)
        			list.insert(tree.lines.peek(i));
			pGetSortedLines(tree.back);
     		}else{   /* the eye is on the partition plane */
            	pGetSortedLines(tree.front);
            	pGetSortedLines(tree.back);
       	}
 	}

 	/**
	 * a function to get the sorted "line" list.
	 * 
	 * @param eye the pPoint2D object representing the eye.
	 * @return a pQueue object of the sorted lines in relation to the eye.  
	 */
   	public pQueue pGetSortedLines(pPoint2D eye)
   	{
     		list = new pQueue();
        	this.eye = new pPoint2D(eye);
       	pGetSortedLines(root);
        	return list;
   	}
	
	/**
	 * a standard <code>toString</code> method, returns the whole
	 * tree in a <code>String</code> object.
	 *
	 * @return the whole tree in a String object.
	 */
	public String toString(){
		list = new pQueue();
		String s = new String();
		if(root == null) return null;
		intrav(root);
		while(!list.empty())
			s += (String)list.remove();
		return (s+="\n");		
	}
	
	/**
	 * a method used by the "toString" to get the list of 
	 * lines in an intrav order. 
	 * 
	 * @param p the current node in recursion
	 */
	public void intrav(pBSPNode p){
		if(p == null) return;
		intrav(p.getback());
		for(int i = 0;i < p.lines.number();i++)
			list.insert(p.lines.peek(i));
		intrav(p.getfront());
	}
}
