/*
 * @(#)pPlane2D.java    Dec 31, 1997
 * 
 * Copyright (c) 1997, Particle
 * 
 * Sale of this "Software" is prohibited!  Only FREE
 * distribution is alowed.
 *
 * -- Particle --
 */

import pPoint2D;
import pLine2D;

/**
 * A <code>pPlane2D</code> class is a very basic 2D Plane.
 * <p>
 * It alows users to set and get values of a 2D Plane, 
 * do some weird stuff with the plane, like split lines
 * wiht it... etc...
 *
 * @version     1.0 12/31/1997
 * @author      Particle
 */

public class pPlane2D{
	/**
	 * the Epsilon value to make this plane "thick"
	 */
  	public final static double EPSILON = 0.001;
      
	/** 
	 * The statics to differentiate between front/back/spanning
	 * "evaluations" of a point...
	 */  
	public final static int SPANNING = 0;
     	public final static int IN_FRONT = 1;
  	public final static int IN_BACK = 2;
     	public final static int COINCIDENT = 3;

	/**
	 * The 2 element array, with first standing for the 
	 * A, second standing for B, and third standing
	 * for C in the:  <code>Ax + By - C = 0</code>
	 * line equation.
	 */
    	public double [] v = null;

	/** 
	 * Constructs a new pPlane2D object
	 */
    	public pPlane2D(){
     		v = new double[3];
    	}
    	
	/**
	 * Constructs a new pPlane2D object given a line.
	 * <p>
	 * let Y1 stand for first y, let Y2 stand for second y, 
	 * and so on...  let FX stand for starting point X, and FY 
	 * stand for starting point Y, and so on... 
	 * than the equation of a line would be:
	 * <p>
	 * A = -DY;
	 * B = DX;
	 * C = DY * FX - DX * FY;
	 * 
	 * @param l the line of which to make the pPlane2D.
	 */
	public pPlane2D(pLine2D l){
      	v = new double[3];
     		double dx = (l.getEnd().getx() - l.getStart().getx());
    		double dy = (l.getEnd().gety() - l.getStart().gety());
    		A(-dy);
    		B(dx);
    		C(dy*l.getStart().getx()-dx*l.getStart().gety());
 	}

    	/**
	 * Constructs a pPlane2D object out of a pPlane2D object,
	 * (basically a copy constructor)
	 * @param l the plane to "copy"
	 */
	public pPlane2D(pPlane2D l){
      	v = new double[3];
       	A(l.A());
       	B(l.B());
         	C(l.C());
   	}

	/**
	 * Sets A in the line equation:
	 * Ax + By - C
	 */
	public void A(double a){
		v[0] = a;
	}

	/**
	 * Sets B in the line equation:
	 * Ax + By - C
	 */
	public void B(double b){
		v[1] = b;
	}

	/**
	 * Sets C in the line equation:
	 * Ax + By - C
	 */
	public void C(double c){
		v[2] = c;
	}

	/**
	 * Gets A in the line equation:
	 * Ax + By - C
	 */
	public double A(){
		return v[0];
	}

	/**
	 * Gets B in the line equation:
	 * Ax + By - C
	 */
	public double B(){
		return v[1];
	}

	/**
	 * Gets C in the line equation:
	 * Ax + By - C
	 */
	public double C(){
		return v[2];
	}

    	/** 
	 * Function to evaluate whether a point (pPoint2D) 
	 * is in "front" or "behind" the plane.
	 * also makes sure the plane is "thick"
	 *
	 * @param p the point to evaluate
	 * @return a condition, one of the statics like IN_FRONT, etc...
	 */
    	public int eval(pPoint2D p){
      	double c = A()*p.getx() + B()*p.gety() + C();
       	if(c > pPlane2D.EPSILON){
    			return pPlane2D.IN_FRONT;
      	}else if(c < (-pPlane2D.EPSILON)){
       		return pPlane2D.IN_BACK;
    		}else{
           		return pPlane2D.SPANNING;
       	}
	}

	/** 
	 * the function to evaluate a line in terms of this plane
	 * 
	 * @return one of the "statics" like IN_FRONT, etc...
	 * @param l the line to evaluate
	 */
	public int eval(pLine2D l){
      	int a = eval(l.getStart());
      	int b = eval(l.getEnd());
      	if(a == pPlane2D.SPANNING)
         		if(b == pPlane2D.SPANNING)
                  	return pPlane2D.COINCIDENT;
           		else return b;
         	if(b == pPlane2D.SPANNING)
            	if(a == pPlane2D.SPANNING)
                  	return pPlane2D.COINCIDENT;
            	else return a;
        	if((a == pPlane2D.IN_FRONT)&&(b == pPlane2D.IN_BACK))
            	return SPANNING;
        	if((a == pPlane2D.IN_BACK)&&(b == pPlane2D.IN_FRONT))
            	return SPANNING;
		return a;   /* all on 1 side */
 	}

	/** 
 	 * function to split a line with this plane.
	 * <p>
	 * the "front" piece of the split becomes the
	 * first element in the returning array, the 
	 * "back" piece of the split, becomes the 
	 * second.
	 * <p>
	 * there is an assumption, that "positive eval()" 
	 * is front, and negative eval() is in back.
	 * @return a 2 element array, with front and back pieces of the line
	 * @param l the line to split.
	 * @see pLine2D
	 * @see pPoint2D
	 */
   	public pLine2D [] split(pLine2D l){
      	pPlane2D p = new pPlane2D(l);
        	pLine2D [] q = new pLine2D[2];
        	double cross_x = 0, cross_y = 0;
        	double divider = A() * p.B() - B() * p.A();
      	if(divider == 0){   /* should never happen */
            	if(p.A() == 0){
                  	cross_x = l.getStart().getx();
             	}
             	if(p.B() == 0){
                  	cross_y = l.getStart().gety();
             	}
            	if(A() == 0){
                   	cross_y = -B();
              	}
             	if(B() == 0){
                 		cross_x = C();
              	}
     		}else{
            	cross_x = (-C()*p.B() + B()*p.C())/divider;
            	cross_y = (-A()*p.C() + C()*p.A())/divider;
       	}           
         	int p1 = eval(l.getStart());
        	int p2 = eval(l.getEnd());
        	q[0] = new pLine2D(l);
         	q[1] = new pLine2D(l);
         	if((p1==pPlane2D.IN_BACK)&&(p2==pPlane2D.IN_FRONT)){
            	q[0].setStart(new pPoint2D(cross_x,cross_y));
          		q[0].setEnd(new pPoint2D(l.getEnd()));
          		q[1].setStart(new pPoint2D(l.getStart()));
           		q[1].setEnd(new pPoint2D(cross_x,cross_y));
      	}else if((p1==pPlane2D.IN_FRONT)&&(p2==pPlane2D.IN_BACK)){
            	q[0].setStart(new pPoint2D(l.getStart()));
            	q[0].setEnd(new pPoint2D(cross_x,cross_y));
           		q[1].setStart(new pPoint2D(cross_x,cross_y));
          		q[1].setEnd(new pPoint2D(l.getEnd()));
      	}else{
            	return null; /* trying to split a spanning line */
       	}
      	return q;
	}

	/**
	 * the standard <code>toString</code> method
	 */
 	public String toString(){
      	return new String("pPlane2D("+A()+","+B()+","+C()+");\n");
   	}
}
