/*
 * @(#)pQueue.java    Dec 31, 1997
 * 
 * Copyright (c) 1997, Particle
 * 
 * Sale of this "Software" is prohibited!  Only FREE
 * distribution is alowed.
 *
 * -- Particle --
 */

import pNode;

/**
 * A <code>pQueue</code> class is a very basic Queue.
 * <p>
 * It alows uses to insert and remove stuff from the queue, 
 * it also alows the ability to "peek" into the queue, and 
 * to find out the number of items on the queue.
 * <p>
 * Definately needs improment!  A "good" queue should have 
 * 2 pointers, to the back and to the front.  Nodes should also 
 * be doubly linked, to to make queue operations of the queue 
 * faster;  as it is, it's the slowest possible implementation.
 *
 * @version     1.0 12/31/1997
 * @author      Particle
 */

public class pQueue{
	
	/**
	 * this is the "header" node
	 */
	protected pNode list;

	/** 
	 * a counter to count how many items are on the queue
	 */
	int counter;

    	/**
	 * Constructs a pQueue, initialized header node to null, and counter to 0.
	 */
	public pQueue(){
          	counter = 0;
         	list = null;
  	}

	/**
	 * returns whether or not the queue is empty.
	 * @return a boolean value of whether or not the queue is empty.
	 */
    	public boolean empty(){
      	return (list == null);
   	}

	/**
	 * gets the number of items in the queue.
	 * @return number of items in queue
	 */
	public int number(){
		return counter;
	}

	/** 
	 * removes everything from the queue.
	 * <p>
	 * actually, thanks to java's gc, all we need to do is set list to null, 
	 * but this method is "more re-ashuring" ;-)
	 */
	public void clear(){
      	while(!empty()) 
			remove();
		counter = 0;
  	}
      
	/** 
	 * a private method to insert an object onto the back of the queue, 
	 * this is not a standard queue operation, thus, it's private.
	 *
	 * @param obj the object to insert
	 */  
    	void insertend(Object obj){
      	counter++;
      	list = new pNode(obj,list);
   	}
      
	/**
	 * removes an object from the queue, a cast will uauslly be needed when
	 * calling this function, to convert <code>Object</code> type into whatever
	 * type it's suppose to be... (an "instance of" can also be used)
	 *
	 * @return the object from the queue.
	 */  
	public Object remove(){
		if(empty()) return null;
          	counter--;
		pNode tmp = new pNode(list.getdata(),list.getnext());
        	list = tmp.getnext();
        	return tmp.getdata();
 	}
	
	/**
         * inserts an <code>Object</code> into the queue.  
	 *
	 * @param obj the object to insert
	 */
  	public void insert(Object obj){
      	if(list == null){ 
			this.insertend(obj);
		}else{
			counter++;
			pNode t = list;
			while(t.getnext() != null)
				t=t.getnext();
			pNode tmp = new pNode(obj,t.getnext());
			t.setnext(tmp);
		}
	}

	/** 
	 * removes an object from the end of the queue, not the 
	 * "correct" way to remove stuff from the queue, thus, it's
	 * private.
	 *
	 * @return the object from the end of the queue
	 */
	Object removeend(){
		if(empty()) return null;
		if(list.getnext() == null) return this.remove();
		counter--;
		pNode t = list;
		while(t.getnext().getnext() != null)
			t = t.getnext();
		Object obj = t.getnext().getdata();
		t.setnext(t.getnext().getnext());
		return obj;
	}

	/** 
	 * peeks into the queue.  You can use this method to 
	 * see what's inside the queue, (searching the queue for 
	 * for something, or just iterating it... 
	 * 
	 * @param pos a "position" (or count) of where to peek.
	 * @return the object at that position, or null on error.
	 */	
  	public Object peek(int peek_pos){
                int i=0;
        	pNode tmp = list;
         	if(tmp == null || (peek_pos < 0)) return null;
          	while((i++ < peek_pos) && (tmp.getnext() != null))
            	tmp = tmp.getnext();
          	return tmp.getdata();
   	}
	
	/**
	 * the standard <code>toString()</code> method.
	 */
	public String toString(){
                pNode n = list;
                String s = new String();
                while(n != null){
            	s+=n.getdata()+"\n";
              	n = n.getnext();
       	}
                s += "\n";
         	return s;
	}
}
